# -*- coding: utf-8 -*-

import os

BASE_URL = "http://pokeapi.co/api/v2"
SPRITE_URL = "https://raw.githubusercontent.com/PokeAPI/sprites/master/sprites"
ENDPOINTS = [
    "ability",
    "berry",
    "berry-firmness",
    "berry-flavor",
    "characteristic",
    "contest-effect",
    "contest-type",
    "egg-group",
    "encounter-condition",
    "encounter-condition-value",
    "encounter-method",
    "evolution-chain",
    "evolution-trigger",
    "gender",
    "generation",
    "growth-rate",
    "item",
    "item-attribute",
    "item-category",
    "item-fling-effect",
    "item-pocket",
    "language",
    "location",
    "location-area",
    "machine",
    "move",
    "move-ailment",
    "move-battle-style",
    "move-category",
    "move-damage-class",
    "move-learn-method",
    "move-target",
    "nature",
    "pal-park-area",
    "pokeathlon-stat",
    "pokedex",
    "pokemon",
    "pokemon-color",
    "pokemon-form",
    "pokemon-habitat",
    "pokemon-shape",
    "pokemon-species",
    "region",
    "stat",
    "super-contest-effect",
    "type",
    "version",
    "version-group",
]
SPRITE_EXT = "png"


def validate(endpoint, resource_id=None):
    if endpoint not in ENDPOINTS:
        raise ValueError("Unknown API endpoint '{}'".format(endpoint))

    if resource_id is not None and not isinstance(resource_id, int):

        raise ValueError("Bad id '{}'".format(resource_id))

    return None


def api_url_build(endpoint, resource_id=None, subresource=None):
    validate(endpoint, resource_id)

    if resource_id is not None:
        if subresource is not None:
            return "/".join([BASE_URL, endpoint, str(resource_id), subresource, ""])

        return "/".join([BASE_URL, endpoint, str(resource_id), ""])

    return "/".join([BASE_URL, endpoint, ""])


def cache_uri_build(endpoint, resource_id=None, subresource=None):
    validate(endpoint, resource_id)

    if resource_id is not None:
        if subresource is not None:
            return "/".join([endpoint, str(resource_id), subresource, ""])

        return "/".join([endpoint, str(resource_id), ""])

    return "/".join([endpoint, ""])


def sprite_url_build(sprite_type, sprite_id, **kwargs):
    options = parse_sprite_options(sprite_type, **kwargs)

    filename = ".".join([str(sprite_id), SPRITE_EXT])
    url = "/".join([SPRITE_URL, sprite_type, *options, filename])

    return url


def sprite_filepath_build(sprite_type, sprite_id, **kwargs):
    """returns the filepath of the sprite *relative to SPRITE_CACHE*"""

    options = parse_sprite_options(sprite_type, **kwargs)

    filename = ".".join([str(sprite_id), SPRITE_EXT])
    filepath = os.path.join(sprite_type, *options, filename)

    return filepath


def parse_sprite_options(sprite_type, **kwargs):
    options = []

    if sprite_type == "pokemon":
        if kwargs.get("model", False):
            options.append("model")
        elif kwargs.get("other", False):
            options.append("other")
            if kwargs.get("official_artwork", False):
                options.append("official-artwork")
            if kwargs.get("dream_world", False):
                options.append("dream-world")
        else:
            if kwargs.get("back", False):
                options.append("back")
            if kwargs.get("shiny", False):
                options.append("shiny")
            if kwargs.get("female", False):
                options.append("female")
    elif sprite_type == "items":
        if kwargs.get("berries", False):
            options.append("berries")
        elif kwargs.get("dream_world", False):
            options.append("dream-world")
        elif kwargs.get("gen3", False):
            options.append("gen3")
        elif kwargs.get("gen5", False):
            options.append("gen5")
        elif kwargs.get("underground", False):
            options.append("underground")

    return options



# -*- coding: utf-8 -*-

from .api import get_data, get_sprite
from .common import api_url_build, sprite_url_build


def _make_obj(obj):
    """Takes an object and returns a corresponding API class.

    The names and values of the data will match exactly with those found
    in the online docs at https://pokeapi.co/docsv2/ . In some cases, the data
    may be of a standard type, such as an integer or string. For those cases,
    the input value is simply returned, unchanged.

    :param obj: the object to be converted
    :return either the same value, if it does not need to be converted, or a
    APIResource or APIMetadata instance, depending on the data inputted.
    """

    def change_sprite_key(d):
        new_dict = {}
        for k, v in d.items():
            if isinstance(v, dict):
                v = change_sprite_key(v)
            new_dict[k.replace('-', '_')] = v
        return new_dict

    if isinstance(obj, dict):
        if "url" in obj.keys():
            url = obj["url"]
            id_ = int(url.split("/")[-2])  # ID of the data.
            endpoint = url.split("/")[-3]  # Where the data is located.
            return APIResource(endpoint, id_, lazy_load=True)
        if all(k in obj for k in ("other", "back_default")):
            obj = change_sprite_key(obj)  # Change hyphens in sprite keys to underscores

        return APIMetadata(obj)

    return obj


def name_id_convert(endpoint, name_or_id):
    if isinstance(name_or_id, int):
        id_ = name_or_id
        name = _convert_id_to_name(endpoint, id_)

    elif isinstance(name_or_id, str):
        name = name_or_id
        id_ = _convert_name_to_id(endpoint, name)

    else:
        raise ValueError(f"the name or id '{name_or_id}' could not be converted")

    return name, id_


def _convert_id_to_name(endpoint, id_):
    resource_data = get_data(endpoint)["results"]

    for resource in resource_data:
        if resource["url"].split("/")[-2] == str(id_):

            # Return the matching name, or id_ if it doesn't exsist.
            return resource.get("name", str(id_))

    return None


def _convert_name_to_id(endpoint, name):
    resource_data = get_data(endpoint)["results"]

    for resource in resource_data:
        if resource.get("name") == name:
            return int(resource.get("url").split("/")[-2])

    return None


class APIResource(object):
    """Core API class, used for accessing the bulk of the data.

    The class uses a modified __getattr__ function to serve the appropriate
    data, so lookup data via the `.` operator, and use the `PokeAPI docs
    <https://pokeapi.co/docsv2/>`_ or the builtin `dir` function to see the
    possible lookups.

    This class takes the complexity out of lots of similar classes for each
    different kind of data served by the API, all of which are very similar,
    but not identical.
    """

    def __init__(
        self, endpoint, name_or_id, lazy_load=False, force_lookup=False, custom=None
    ):

        name, id_ = name_id_convert(endpoint, name_or_id)
        url = api_url_build(endpoint, id_)

        self.__dict__.update({"name": name, "endpoint": endpoint, "id_": id_, "url": url})

        self.__loaded = False
        self.__force_lookup = force_lookup

        if custom:
            self._custom = custom
        else:
            self._custom = {}

        if not lazy_load:
            self._load()
            self.__loaded = True

    def __getattr__(self, attr):
        """Modified method to auto-load the data when it is needed.

        If the data has not yet been looked up, it is loaded, and then checked
        for the requested attribute. If it is not found, AttributeError is
        raised.
        """

        if not self.__loaded:
            self._load()
            self.__loaded = True

            return self.__getattribute__(attr)

        else:
            raise AttributeError(f"{type(self)} object has no attribute {attr}")

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        return f"<{self.endpoint}-{self.name}>"

    def _load(self):
        """Function to collect reference data and connect it to the instance as
         attributes.

         Internal function, does not usually need to be called by the user, as
         it is called automatically when an attribute is requested.

        :return None
        """

        data = get_data(self.endpoint, self.id_, force_lookup=self.__force_lookup)

        # Make our custom objects from the data.
        for key, val in data.items():
            if key in self._custom:
                val = get_data(*self._custom[key](val))

            if isinstance(val, dict):
                data[key] = _make_obj(val)

            elif isinstance(val, list):
                data[key] = [_make_obj(i) for i in val]

        self.__dict__.update(data)

        return None


class APIResourceList(object):
    """Class for a data container.

    Used to access data corresponding to a category, rather than an individual
    reference. Ex. APIResourceList('berry') gives information about all
    berries, such as which ID's correspond to which berry names, and
    how many berries there are.

    You can iterate through all the names or all the urls, using the respective
    properties. You can also iterate on the object itself to run through the
    `dict`s with names and urls together, whatever floats your boat.
    """

    def __init__(self, endpoint, force_lookup=False):
        """Creates a new APIResourceList instance.

        :param name: the name of the resource to get (ex. 'berry' or 'move')
        """

        response = get_data(endpoint, force_lookup=force_lookup)

        self.name = endpoint
        self.__results = [i for i in response["results"]]
        self.count = response["count"]

    def __len__(self):
        return self.count

    def __iter__(self):
        return iter(self.__results)

    def __str__(self):
        return str(self.__results)

    @property
    def names(self):
        """Useful iterator for all the resource's names."""
        for result in self.__results:
            yield result.get("name", result["url"].split("/")[-2])

    @property
    def urls(self):
        """Useful iterator for all of the resource's urls."""
        for result in self.__results:
            yield result["url"]


class APIMetadata(object):
    """Helper class for smaller references.

    This class emulates a dictionary, but attribute lookup is via the `.`
    operator, not indexing. (ex. instance.attr, not instance['attr']).

    Used for "Common Models" classes and APIResource helper classes.
    https://pokeapi.co/docsv2/#common-models
    """

    def __init__(self, data):

        for key, val in data.items():

            if isinstance(val, dict):
                data[key] = _make_obj(val)

            if isinstance(val, list):
                data[key] = [_make_obj(i) for i in val]

        self.__dict__.update(data)


class SpriteResource(object):
    def __init__(self, sprite_type, sprite_id, **kwargs):

        url = sprite_url_build(sprite_type, sprite_id, **kwargs)

        self.__dict__.update(
            {"sprite_id": sprite_id, "sprite_type": sprite_type, "url": url}
        )

        self.__loaded = False
        self.__force_lookup = kwargs.get("force_lookup", False)
        self.__orginal_kwargs = kwargs

        if not kwargs.get("lazy_load", False):
            self._load()
            self.__loaded = True

    def _load(self):
        data = get_sprite(self.sprite_type, self.sprite_id, **self.__orginal_kwargs)
        self.__dict__.update(data)

        return None

    def __getattr__(self, attr):
        """Modified method to auto-load the data when it is needed.

        If the data has not yet been looked up, it is loaded, and then checked
        for the requested attribute. If it is not found, AttributeError is
        raised.
        """

        if not self.__loaded:
            self._load()
            self.__loaded = True

            return self.__getattribute__(attr)

        else:
            raise AttributeError(f"{type(self)} object has no attribute {attr}")



# -*- coding: utf-8 -*-

import importlib
import os
import shelve
import unittest

from hypothesis import assume, given
from hypothesis.strategies import characters, dictionaries, integers, sampled_from, text

from pokebase import cache
from pokebase.common import ENDPOINTS


class TestFunction_save(unittest.TestCase):

    # cache.save(data, endpoint, resource_id=None)

    def setUp(self):
        cache.set_cache('testing')

    @given(data=dictionaries(text(), text()),
           endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1))
    def testArgs(self, data, endpoint, resource_id):
        self.assertIsNone(cache.save(data, endpoint, resource_id))

    @given(data=text(),
           endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1))
    def testArg_data_Text(self, data, endpoint, resource_id):
        with self.assertRaises(ValueError):
            cache.save(data, endpoint, resource_id)

    @given(data=dictionaries(text(), text()),
           endpoint=text(),
           resource_id=integers(min_value=1))
    def testArg_endpoint_Text(self, data, endpoint, resource_id):
        assume(data != dict())
        with self.assertRaises(ValueError):
            cache.save(data, endpoint, resource_id)

    @given(data=dictionaries(text(), text()),
           endpoint=sampled_from(ENDPOINTS),
           resource_id=text())
    def testArg_resource_id_Text(self, data, endpoint, resource_id):
        assume(data != dict())
        with self.assertRaises(ValueError):
            cache.save(data, endpoint, resource_id)

    @given(data=dictionaries(text(), text()),
           endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1),
           subresource=text())
    def testArg_subresource_Text(self, data, endpoint, resource_id, subresource):
        self.assertIsNone(cache.save(data, endpoint, resource_id, subresource))

    @given(data=dictionaries(text(), text()),
           endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1))
    def testEnv_CacheFileNotFound(self, data, endpoint, resource_id):
        assume(data != dict())
        os.remove(cache.API_CACHE)
        self.assertIsNone(cache.save(data, endpoint, resource_id))

    @given(data=dictionaries(text(), text()),
           endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1))
    def testEnd_CacheFileAlreadyOpen(self, data, endpoint, resource_id):
        cache_db = shelve.open(cache.API_CACHE)
        self.assertIsNone(cache.save(data, endpoint, resource_id))
        cache_db.close()


class TestFunction_load(unittest.TestCase):

    # cache.load(endpoint, resource_id=None)

    def setUp(self):
        cache.set_cache('testing')

    @given(data=dictionaries(text(), text()),
           endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1))
    def testArgs(self, data, endpoint, resource_id):
        assume(data != dict())
        cache.save(data, endpoint, resource_id)
        self.assertEqual(data, cache.load(endpoint, resource_id))

    @given(endpoint=text(),
           resource_id=integers(min_value=1))
    def testArg_endpoint_Text(self, endpoint, resource_id):
        with self.assertRaises(ValueError):
            cache.load(endpoint, resource_id)

    @given(endpoint=sampled_from(ENDPOINTS),
           resource_id=text())
    def testArg_resource_id_Text(self, endpoint, resource_id):
        with self.assertRaises(ValueError):
            cache.load(endpoint, resource_id)

    @given(data=dictionaries(text(), text()),
           endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1),
           subresource=text())
    def testArg_subresource_Text(self, data, endpoint, resource_id, subresource):
        assume(data != dict())
        cache.save(data, endpoint, resource_id, subresource)
        self.assertEqual(data, cache.load(endpoint, resource_id, subresource))

    @given(endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1))
    def testEnv_CacheFileNotFound(self, endpoint, resource_id):
        # ensure it exsists before we delete it,
        cache.set_cache('testing')
        os.remove(cache.API_CACHE)
        with self.assertRaises(KeyError):
            cache.load(endpoint, resource_id)

    @given(endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1))
    @unittest.skip('inconsistency between Travis-CI/local machine')
    def testEnv_CacheFileAlreadyOpen(self, endpoint, resource_id):
        cache_db = shelve.open(cache.API_CACHE)
        with self.assertRaises(KeyError):
            cache.load(endpoint, resource_id)
        cache_db.close()

    @given(endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1))
    def testEnv_KeyNotInCache(self, endpoint, resource_id):

        with shelve.open(cache.API_CACHE) as c:
            key = cache.cache_uri_build(endpoint, resource_id)
            if key in c:
                del c[key]

        with self.assertRaises(KeyError):
            cache.load(endpoint, resource_id)


class TestFunction_set_cache(unittest.TestCase):

    # cache.set_cache(new_path=None)

    default_home = os.path.join(os.path.expanduser('~'), '.cache')

    def testAttr_Caches_Import(self):
        importlib.reload(cache)
        self.assertEqual(os.path.join(self.default_home, 'pokebase'),
                         cache.get_default_cache())
        self.assertEqual(os.path.join(self.default_home, 'pokebase'),
                         cache.CACHE_DIR)
        self.assertEqual(os.path.join(self.default_home, 'pokebase', 'api.cache'),
                         cache.API_CACHE)
        self.assertEqual(os.path.join(self.default_home, 'pokebase', 'sprite'),
                         cache.SPRITE_CACHE)

    def testAttr_Caches_Default(self):
        cache_dir, api_cache, sprite_cache = cache.set_cache()
        self.assertEqual(os.path.join(self.default_home, 'pokebase'),
                         cache_dir)
        self.assertEqual(os.path.join(self.default_home, 'pokebase', 'api.cache'),
                         api_cache)
        self.assertEqual(os.path.join(self.default_home, 'pokebase', 'sprite'),
                         sprite_cache)

    def testEnv_CacheDirNotFound(self):
        cache.set_cache('testing')
        os.rmdir(cache.SPRITE_CACHE)
        if os.path.exists(cache.API_CACHE): os.remove(cache.API_CACHE)
        os.rmdir(cache.CACHE_DIR)
        self.assertEqual(cache.set_cache(),
                         (cache.CACHE_DIR, cache.API_CACHE, cache.SPRITE_CACHE))

    @given(new_path=characters(whitelist_categories=['Lu', 'Ll', 'Nd']))
    def testAttr_Caches_PathsChanged(self, new_path):
        cache.set_cache(new_path)
        self.assertEqual(os.path.abspath(new_path),
                         cache.CACHE_DIR)
        self.assertEqual(os.path.join(os.path.abspath(new_path), 'api.cache'),
                         cache.API_CACHE)
        self.assertEqual(os.path.join(os.path.abspath(new_path), 'sprite'),
                         cache.SPRITE_CACHE)
        os.rmdir(cache.SPRITE_CACHE)
        os.rmdir(cache.CACHE_DIR)



# -*- coding: utf-8 -*-

import unittest

from hypothesis import given
from hypothesis.strategies import integers, none, sampled_from, text

from pokebase import common


class TestFunction_validate(unittest.TestCase):

    @given(endpoint=sampled_from(common.ENDPOINTS),
           resource_id=none())
    def testArg_endpoint_Sampled(self, endpoint, resource_id):
        self.assertIsNone(common.validate(endpoint, resource_id))

    @given(endpoint=sampled_from(common.ENDPOINTS),
           resource_id=integers(min_value=1))
    def testArg_resource_id_NonNegInt(self, endpoint, resource_id):
        self.assertIsNone(common.validate(endpoint, resource_id))

    @given(endpoint=text(),
           resource_id=none())
    def testArg_endpoint_Text(self, endpoint, resource_id):
        with self.assertRaises(ValueError):
            common.validate(endpoint, resource_id)

    @given(endpoint=sampled_from(common.ENDPOINTS),
           resource_id=text())
    def testArg_resource_id_Text(self, endpoint, resource_id):
        with self.assertRaises(ValueError):
            common.validate(endpoint, resource_id)


class TestFunction_api_uri_build(unittest.TestCase):

    @given(endpoint=sampled_from(common.ENDPOINTS),
           resource_id=integers(min_value=1))
    def testArg_resource_id_NonNegInt(self, endpoint, resource_id):
        self.assertEqual(common.api_url_build(endpoint, resource_id),
                         'http://pokeapi.co/api/v2/{}/{}/'
                         .format(endpoint, resource_id))

    @given(endpoint=sampled_from(common.ENDPOINTS),
           resource_id=none())
    def testArg_resource_id_None(self, endpoint, resource_id):
        self.assertEqual(common.api_url_build(endpoint, resource_id),
                         'http://pokeapi.co/api/v2/{}/'
                         .format(endpoint))

    @given(endpoint=text(),
           resource_id=none())
    def testArg_endpoint_Text(self, endpoint, resource_id):
        with self.assertRaises(ValueError):
            common.api_url_build(endpoint, resource_id)

    @given(endpoint=sampled_from(common.ENDPOINTS),
           resource_id=integers(min_value=1),
           subresource=text())
    def testArg_subresource_Text(self, endpoint, resource_id, subresource):
        self.assertEqual(common.api_url_build(endpoint, resource_id, subresource),
                         'http://pokeapi.co/api/v2/{}/{}/{}/'
                         .format(endpoint, resource_id, subresource))


class TestFunction_cache_uri_build(unittest.TestCase):

    @given(endpoint=sampled_from(common.ENDPOINTS),
           resource_id=integers(min_value=1))
    def testArgs(self, endpoint, resource_id):
        self.assertEqual(common.cache_uri_build(endpoint, resource_id),
                         '{}/{}/'.format(endpoint, resource_id))

    @given(endpoint=sampled_from(common.ENDPOINTS),
           resource_id=none())
    def testArg_resource_id_None(self, endpoint, resource_id):
        self.assertEqual(common.cache_uri_build(endpoint, resource_id),
                         '/'.join([endpoint, '']))

    @given(endpoint=text(),
           resource_id=none())
    def testArg_endpoint_Text(self, endpoint, resource_id):
        with self.assertRaises(ValueError):
            common.cache_uri_build(endpoint, resource_id)

    @given(endpoint=sampled_from(common.ENDPOINTS),
           resource_id=integers(min_value=1),
           subresource=text())
    def testArg_subresource_Text(self, endpoint, resource_id, subresource):
        self.assertEqual(common.cache_uri_build(endpoint, resource_id, subresource),
                         '/'.join([endpoint, str(resource_id), subresource, '']))



# -*- coding: utf-8 -*-

import unittest
from os import environ

import pokebase as pb


@unittest.skipIf(environ.get('MOCK_ONLY', False), 'only running mock API calls')
class TestAPICalls(unittest.TestCase):

    def testFunction__call_api(self):
        self.assertIsInstance(pb.api._call_api('berry', 1),
                              dict)

    def testFunction_get_data(self):
        self.assertIsInstance(pb.api.get_data('berry', 1, force_lookup=True),
                              dict)

    def testClass_APIResource(self):
        self.assertIsInstance(pb.APIResource('berry', 1, force_lookup=True),
                              pb.APIResource)

    def testClass_APIResourceList(self):
        self.assertIsInstance(pb.APIResourceList('berry', force_lookup=True),
                              pb.APIResourceList)



# -*- coding: utf-8 -*-

import unittest
from unittest.mock import patch

from hypothesis import given
from hypothesis.strategies import dictionaries, integers, lists, sampled_from, text

from pokebase import interface
from pokebase.cache import set_cache
from pokebase.common import ENDPOINTS
from pokebase.loaders import pokemon


class TestFunction__make_obj(unittest.TestCase):

    # _make_obj(obj)

    def setUp(self):
        set_cache('testing')

    @given(obj=dictionaries(text(), text()))
    def testArg_obj_Dictionary(self, obj):
        self.assertIsInstance(interface._make_obj(obj), interface.APIMetadata)

    @given(endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1),
           obj=dictionaries(text(), text()))
    @patch('pokebase.interface.get_data')
    def testArg_obj_DictionaryWithUrl(self, mock_get_data, endpoint, resource_id, obj):

        mock_get_data.return_value = {'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, resource_id)}]}

        obj['url'] = 'http://base.url/{}/{}/'.format(endpoint, resource_id)
        self.assertIsInstance(interface._make_obj(obj), interface.APIResource)

    @given(obj=lists(elements=text()))
    def testArg_obj_List(self, obj):
        self.assertEqual(obj, interface._make_obj(obj))

    @given(obj=text())
    def testArg_obj_Str(self, obj):
        self.assertEqual(obj, interface._make_obj(obj))


class TestFunction__convert_id_to_name(unittest.TestCase):

    # _convert_id_to_name(endpoint, id_)

    def setUp(self):
        set_cache('testing')

    @given(name=text(),
           endpoint=sampled_from(ENDPOINTS),
           id_=integers(min_value=1))
    @patch('pokebase.interface.get_data')
    def testArgs(self, mock_get_data, name, endpoint, id_):

        mock_get_data.return_value = {'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, id_), 'name': name}]}

        self.assertEqual(name, interface._convert_id_to_name(endpoint, id_))

    @given(endpoint=sampled_from(ENDPOINTS),
           id_=integers(min_value=1))
    @patch('pokebase.interface.get_data')
    def testEnv_NotNamedResource(self, mock_get_data, endpoint, id_):

        mock_get_data.return_value = {'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, id_)}]}

        self.assertEqual(str(id_), interface._convert_id_to_name(endpoint, id_))

    @given(endpoint=text(),
           id_=integers(min_value=1))
    @patch('pokebase.interface.get_data')
    def testArg_endpoint_Text(self, mock_get_data, endpoint, id_):

        mock_get_data.side_effect = ValueError()

        with self.assertRaises(ValueError):
            interface._convert_id_to_name(endpoint, id_)


class TestFunction__convert_name_to_id(unittest.TestCase):

    # _convert_name_to_id(endpoint, name)
    @given(id_=integers(min_value=1),
           endpoint=sampled_from(ENDPOINTS),
           name=text())
    @patch('pokebase.interface.get_data')
    def testArgs(self, mock_get_data, id_, endpoint, name):

        mock_get_data.return_value = {'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, id_), 'name': name}]}

        self.assertEqual(id_, interface._convert_name_to_id(endpoint, name))

    @given(endpoint=text(),
           name=text())
    @patch('pokebase.interface.get_data')
    def testArg_endpoint_Text(self, mock_get_data, endpoint, name):

        mock_get_data.side_effect = ValueError()

        with self.assertRaises(ValueError):
            interface._convert_id_to_name(endpoint, name)


class TestFunction_name_id_convert(unittest.TestCase):

    # name_id_convert(endpont, name_or_id)

    @given(name=text(),
           endpoint=sampled_from(ENDPOINTS),
           id_=integers(min_value=1))
    @patch('pokebase.interface.get_data')
    def testArgs_WithID(self, mock_get_data, name, endpoint, id_):

        mock_get_data.return_value = {'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, id_), 'name': name}]}

        self.assertEqual((name, id_), interface.name_id_convert(endpoint, id_))

    @given(id_=integers(min_value=1),
           endpoint=sampled_from(ENDPOINTS),
           name=text())
    @patch('pokebase.interface.get_data')
    def testArgs_WithName(self, mock_get_data, id_, endpoint, name):

        mock_get_data.return_value = {'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, id_), 'name': name}]}

        self.assertEqual((name, id_), interface.name_id_convert(endpoint, id_))

    @given(endpoint=text(),
           name=text())
    @patch('pokebase.interface.get_data')
    def testArg_endpoint_Text(self, mock_get_data, endpoint, name):

        mock_get_data.side_effect = ValueError()

        with self.assertRaises(ValueError):
            interface.name_id_convert(endpoint, name)


class TestClass_APIResource(unittest.TestCase):

    # APIResource(endpoint, name_or_id, lazy_load=True)

    def setUp(self):
        set_cache('testing')

    @given(id_=integers(min_value=1),
           endpoint=sampled_from(ENDPOINTS),
           name=text())
    @patch('pokebase.interface.get_data')
    def testArgs_WithNameWithLazyLoad(self, mock_get_data, id_, endpoint, name):

        mock_get_data.return_value = {'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, id_), 'name': name}]}

        self.assertIsInstance(interface.APIResource(endpoint, name),
                              interface.APIResource)

    @given(name=text(),
           endpoint=sampled_from(ENDPOINTS),
           id_=integers(min_value=1))
    @patch('pokebase.interface.get_data')
    def testArgs_WithIDWithLazyLoad(self, mock_get_data, name, endpoint, id_):

        mock_get_data.return_value = {'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, id_), 'name': name}]}

        self.assertIsInstance(interface.APIResource(endpoint, id_),
                              interface.APIResource)

    @given(id_=integers(min_value=1),
           endpoint=sampled_from(ENDPOINTS),
           name=text())
    @patch('pokebase.interface.get_data')
    def testAttrs_WithNameWithoutLazyLoad(self, mock_get_data, id_, endpoint, name):

        mock_get_data.side_effect = [{'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, id_), 'name': name}]},
                                     {'simple_attr': 10, 'list_attr': [{'name': 'mocked name'}]}]

        sample = interface.APIResource(endpoint, name, lazy_load=False)

        self.assertEqual(sample.simple_attr, 10)
        self.assertIsInstance(sample.list_attr, list)
        self.assertIsInstance(sample.list_attr[0], interface.APIMetadata)

    @given(name=text(),
           endpoint=sampled_from(ENDPOINTS),
           id_=integers(min_value=1))
    @patch('pokebase.interface.get_data')
    def testAttrs_WithIDWithoutLazyLoad(self, mock_get_data, name, endpoint, id_):

        mock_get_data.side_effect = [{'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, id_), 'name': name}]},
                                     {'simple_attr': 10, 'list_attr': [{'name': 'mocked name'}], 'complex_attr': {'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, 10)}},
                                     {'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, id_), 'name': name + '2'}]}]

        sample = interface.APIResource(endpoint, id_, lazy_load=False)

        self.assertEqual(sample.simple_attr, 10)
        self.assertIsInstance(sample.list_attr, list)
        self.assertIsInstance(sample.list_attr[0], interface.APIMetadata)
        self.assertIsInstance(sample.complex_attr, interface.APIResource)

    @given(name=text(),
           endpoint=sampled_from(ENDPOINTS),
           id_=integers(min_value=1))
    @patch('pokebase.interface.get_data')
    def testAttrs_WithLazyLoad(self, mock_get_data, name, endpoint, id_):
        mock_get_data.side_effect = [{'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, id_), 'name': name}]},
                                     {'simple_attr': 10, 'list_attr': [{'name': 'mocked name'}]}]

        sample = interface.APIResource(endpoint, id_, lazy_load=True)
        lazy_len = len(dir(sample))
        sample._load()
        loaded_len = len(dir(sample))

        self.assertGreater(loaded_len, lazy_len)

    @given(endpoint=text(),
           id_=integers(min_value=1))
    @patch('pokebase.interface.get_data')
    def testArg_endpoint_Text(self, mock_get_data, endpoint, id_):

        mock_get_data.return_value = {'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(endpoint, id_)}]}

        with self.assertRaises(ValueError):
            interface.APIResource(endpoint, id_)

    @given(id_=integers(min_value=1))
    @patch('pokebase.interface.get_data')
    def testHasLocationAreaEncounters(self, mock_get_data, id_):

        mock_get_data.side_effect = [{'count': 1, 'results': [{'url': 'mocked.url/api/v2/pokemon/{}/'.format(id_), 'name': 'mocked name'}]},
                                     {'location_area_encounters': '/api/v2/pokemon/{}/encounters'.format(id_)},
                                     [{'version_details': [], 'location_area': {'url': 'mocked.url/api/v2/location-area/1/', 'name': 'mocked location_area'}}],
                                     {'count': 1, 'results': [{'url': 'mocked.url/api/v2/location-area/1/', 'name': 'mocked name'}]}]
        pkmn = pokemon(id_)
        # Should be list or empty list
        self.assertIsInstance(pkmn.location_area_encounters, list)
        # would be str if it was not handled
        self.assertNotIsInstance(pkmn.location_area_encounters, str)


class TestClass_APIResourceList(unittest.TestCase):

    @given(endpoint=sampled_from(ENDPOINTS))
    @patch('pokebase.interface.get_data')
    def testArgs(self, mock_get_data, endpoint):

        mock_get_data.return_value = {'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/1/'.format(endpoint)}]}

        self.assertIsInstance(interface.APIResourceList(endpoint),
                              interface.APIResourceList)

    @given(endpoint=text())
    @patch('pokebase.interface.get_data')
    def testArg_endpoint_Text(self, mock_get_data, endpoint):

        mock_get_data.side_effect = ValueError()

        with self.assertRaises(ValueError):
            interface.APIResourceList(endpoint)


class TestClass_APIMetadata(unittest.TestCase):

    @given(data=dictionaries(text(), text()))
    def testArgs(self, data):
        self.assertIsInstance(interface.APIMetadata(data),
                              interface.APIMetadata)



# -*- coding: utf-8 -*-

import os
import shelve

from .common import cache_uri_build, sprite_filepath_build

# Cache locations will be set at the end of this file.
CACHE_DIR = None
API_CACHE = None
SPRITE_CACHE = None


def save(data, endpoint, resource_id=None, subresource=None):

    if data == dict():  # No point in saving empty data.
        return None

    if not isinstance(data, (dict, list)):
        raise ValueError("Could not save non-dict data")

    uri = cache_uri_build(endpoint, resource_id, subresource)

    try:
        with shelve.open(API_CACHE) as cache:
            cache[uri] = data
    except OSError as error:
        if error.errno == 11:  # Cache open by another person/program
            # print('Cache unavailable, skipping save')
            pass
        else:
            raise error

    return None


def save_sprite(data, sprite_type, sprite_id, **kwargs):

    abs_path = data["path"]

    # Make intermediate directories; this line removes the file+extension.
    dirs = abs_path.rpartition(os.path.sep)[0]
    safe_make_dirs(dirs)

    with open(abs_path, "wb") as img_file:
        img_file.write(data["img_data"])

    return None


def load(endpoint, resource_id=None, subresource=None):

    uri = cache_uri_build(endpoint, resource_id, subresource)

    try:
        with shelve.open(API_CACHE) as cache:
            return cache[uri]
    except OSError as error:
        if error.errno == 11:
            # Cache open by another person/program
            # print('Cache unavailable, skipping load')
            raise KeyError("Cache could not be opened.")
        else:
            raise


def load_sprite(sprite_type, sprite_id, **kwargs):
    abs_path = get_sprite_path(sprite_type, sprite_id, **kwargs)

    with open(abs_path, "rb") as img_file:
        img_data = img_file.read()

    return dict(img_data=img_data, path=abs_path)


def safe_make_dirs(path, mode=0o777):
    """Create a leaf directory and all intermediate ones in a safe way.

    A wrapper to os.makedirs() that handles existing leaf directories while
    avoiding os.path.exists() race conditions.

    :param path: relative or absolute directory tree to create
    :param mode: directory permissions in octal
    :return: The newly-created path
    """
    try:
        os.makedirs(path, mode)
    except OSError as error:
        if error.errno != 17:  # File exists
            raise

    return path


def get_default_cache():
    """Get the default cache location.

    Adheres to the XDG Base Directory specification, as described in
    https://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html

    :return: the default cache directory absolute path
    """

    xdg_cache_home = os.environ.get("XDG_CACHE_HOME") or os.path.join(
        os.path.expanduser("~"), ".cache"
    )

    return os.path.join(xdg_cache_home, "pokebase")


def get_sprite_path(sprite_type, sprite_id, **kwargs):
    rel_filepath = sprite_filepath_build(sprite_type, sprite_id, **kwargs)
    abs_path = os.path.join(SPRITE_CACHE, rel_filepath)

    return abs_path


def set_cache(new_path=None):
    """Simple function to change the cache location.

    `new_path` can be an absolute or relative path. If the directory does not
    exist yet, this function will create it. If None it will set the cache to
    the default cache directory.

    If you are going to change the cache directory, this function should be
    called at the top of your script, before you make any calls to the API.
    This is to avoid duplicate files and excess API calls.

    :param new_path: relative or absolute path to the desired new cache
    directory
    :return: str, str
    """

    global CACHE_DIR, API_CACHE, SPRITE_CACHE

    if new_path is None:
        new_path = get_default_cache()

    CACHE_DIR = safe_make_dirs(os.path.abspath(new_path))
    API_CACHE = os.path.join(CACHE_DIR, "api.cache")
    SPRITE_CACHE = safe_make_dirs(os.path.join(CACHE_DIR, "sprite"))

    return CACHE_DIR, API_CACHE, SPRITE_CACHE


CACHE_DIR, API_CACHE, SPRITE_CACHE = set_cache()



# -*- coding: utf-8 -*-

import unittest
from unittest.mock import patch

from hypothesis import given
from hypothesis.strategies import integers

from pokebase import APIResource, loaders
from pokebase.common import ENDPOINTS


def builder(func, func_name):

    @given(id_=integers(min_value=1))
    @patch('pokebase.interface.get_data')
    def test(self, mock_get_data, id_):
        mock_get_data.side_effect = [{'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(func_name, id_)}]},
                                {'simple_attr': 10, 'list_attr': [{'name': 'mocked name'}], 'complex_attr': {'url': 'mocked.url/api/v2/{}/{}/'.format(func_name, 10)}},
                                {'count': 1, 'results': [{'url': 'mocked.url/api/v2/{}/{}/'.format(func_name, id_)}]}]
        self.assertIsInstance(func(id_), APIResource)

    return test


class TestFunctions_loaders(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        for endpoint in ENDPOINTS:
            if endpoint in ['type']:
                # special cases, need trailing underscore
                func_name = ''.join([endpoint.replace('-', '_'), '_'])
            else:
                func_name = endpoint.replace('-', '_')

            func = getattr(loaders, func_name)

            setattr(cls, 'testLoader_{}'.format(func_name), builder(func, endpoint))

TestFunctions_loaders.setUpClass()



# -*- coding: utf-8 -*-

import requests

from .cache import get_sprite_path, load, load_sprite, save, save_sprite
from .common import api_url_build, sprite_url_build


def _call_api(endpoint, resource_id=None, subresource=None):
    url = api_url_build(endpoint, resource_id, subresource)

    # Get a list of resources at the endpoint, if no resource_id is given.
    get_endpoint_list = resource_id is None

    response = requests.get(url)
    response.raise_for_status()

    data = response.json()

    if get_endpoint_list and data["count"] != len(data["results"]):
        # We got a section of all results; we want ALL of them.
        items = data["count"]
        num_items = dict(limit=items)

        response = requests.get(url, params=num_items)
        response.raise_for_status()

        data = response.json()

    return data


def get_data(endpoint, resource_id=None, subresource=None, **kwargs):
    if not kwargs.get("force_lookup", False):
        try:
            data = load(endpoint, resource_id, subresource)
            return data
        except KeyError:
            pass

    data = _call_api(endpoint, resource_id, subresource)
    save(data, endpoint, resource_id, subresource)

    return data


def _call_sprite_api(sprite_type, sprite_id, **kwargs):
    url = sprite_url_build(sprite_type, sprite_id, **kwargs)

    response = requests.get(url)
    response.raise_for_status()

    abs_path = get_sprite_path(sprite_type, sprite_id, **kwargs)
    data = dict(img_data=response.content, path=abs_path)

    return data


def get_sprite(sprite_type, sprite_id, **kwargs):
    if not kwargs.get("force_lookup", False):
        try:
            data = load_sprite(sprite_type, sprite_id, **kwargs)
            return data
        except FileNotFoundError:
            pass

    data = _call_sprite_api(sprite_type, sprite_id, **kwargs)
    save_sprite(data, sprite_type, sprite_id, **kwargs)

    return data



# -*- coding: utf-8 -*-

from .interface import APIResource, SpriteResource


def berry(id_or_name, **kwargs):
    """Quick berry lookup.

    See https://pokeapi.co/docsv2/#berries for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("berry", id_or_name, **kwargs)


def berry_firmness(id_or_name, **kwargs):
    """Quick berry-firmness lookup.

    See https://pokeapi.co/docsv2/#berry-firmnesses for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("berry-firmness", id_or_name, **kwargs)


def berry_flavor(id_or_name, **kwargs):
    """Quick berry-flavor lookup.

    See https://pokeapi.co/docsv2/#berry-flavors for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("berry-flavor", id_or_name, **kwargs)


def contest_type(id_or_name, **kwargs):
    """Quick contest-type lookup.

    See https://pokeapi.co/docsv2/#contest-types for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("contest-type", id_or_name, **kwargs)


def contest_effect(id_, **kwargs):
    """Quick contest-effect lookup.

    See https://pokeapi.co/docsv2/#contest-effects for attributes and more
    detailed information.

    :param id_: id of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("contest-effect", id_, **kwargs)


def super_contest_effect(id_, **kwargs):
    """Quick super-contest-effect lookup.

    See https://pokeapi.co/docsv2/#super-contest-effects for attributes and
    more detailed information.

    :param id_: id of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("super-contest-effect", id_, **kwargs)


def encounter_method(id_or_name, **kwargs):
    """Quick encounter-method lookup.

    See https://pokeapi.co/docsv2/#encounter-methods for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("encounter-method", id_or_name, **kwargs)


def encounter_condition(id_or_name, **kwargs):
    """Quick encounter-condition lookup.

    See https://pokeapi.co/docsv2/#encounter-conditions for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("encounter-condition", id_or_name, **kwargs)


def encounter_condition_value(id_or_name, **kwargs):
    """Quick encounter-condition-value lookup.

    See https://pokeapi.co/docsv2/#encounter-condition-values for attributes
    and more detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("encounter-condition-value", id_or_name, **kwargs)


def evolution_chain(id_, **kwargs):
    """Quick evolution-chain lookup.

    See https://pokeapi.co/docsv2/#evolution-chains for attributes and more
    detailed information.

    :param id_: id of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("evolution-chain", id_, **kwargs)


def evolution_trigger(id_or_name, **kwargs):
    """Quick evolution-trigger lookup.

    See https://pokeapi.co/docsv2/#evolution-triggers for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("evolution-trigger", id_or_name, **kwargs)


def generation(id_or_name, **kwargs):
    """Quick generation lookup.

    See https://pokeapi.co/docsv2/#generations for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("generation", id_or_name, **kwargs)


def pokedex(id_or_name, **kwargs):
    """Quick pokedex lookup.

    See https://pokeapi.co/docsv2/#pokedexes for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("pokedex", id_or_name, **kwargs)


def version(id_or_name, **kwargs):
    """Quick version lookup.

    See https://pokeapi.co/docsv2/#versions for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("version", id_or_name, **kwargs)


def version_group(id_or_name, **kwargs):
    """Quick version-group lookup.

    See https://pokeapi.co/docsv2/#version-groups for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("version-group", id_or_name, **kwargs)


def item(id_or_name, **kwargs):
    """Quick item lookup.

    See https://pokeapi.co/docsv2/#items for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("item", id_or_name, **kwargs)


def item_attribute(id_or_name, **kwargs):
    """Quick item-attribute lookup.

    See https://pokeapi.co/docsv2/#item-attributes for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("item-attribute", id_or_name, **kwargs)


def item_category(id_or_name, **kwargs):
    """Quick item-category lookup.

    See https://pokeapi.co/docsv2/#item-categories for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("item-category", id_or_name, **kwargs)


def item_fling_effect(id_or_name, **kwargs):
    """Quick item-fling-effect lookup.

    See https://pokeapi.co/docsv2/#item-fling-effects for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("item-fling-effect", id_or_name, **kwargs)


def item_pocket(id_or_name, **kwargs):
    """Quick item-pocket lookup.

    See https://pokeapi.co/docsv2/#item-pockets for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("item-pocket", id_or_name, **kwargs)


def machine(id_, **kwargs):
    """Quick machine lookup.

    See https://pokeapi.co/docsv2/#machines for attributes and more detailed
    information.

    :param id_: id of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("machine", id_, **kwargs)


def move(id_or_name, **kwargs):
    """Quick move lookup.

    See https://pokeapi.co/docsv2/#moves for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("move", id_or_name, **kwargs)


def move_ailment(id_or_name, **kwargs):
    """Quick move-ailment lookup.

    See https://pokeapi.co/docsv2/#move-ailments for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("move-ailment", id_or_name, **kwargs)


def move_battle_style(id_or_name, **kwargs):
    """Quick move-battle-style lookup.

    See https://pokeapi.co/docsv2/#move-battle-styles for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("move-battle-style", id_or_name, **kwargs)


def move_category(id_or_name, **kwargs):
    """Quick move-category lookup.

    See https://pokeapi.co/docsv2/#move-categories for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("move-category", id_or_name, **kwargs)


def move_damage_class(id_or_name, **kwargs):
    """Quick move-damage-class lookup.

    See https://pokeapi.co/docsv2/#move-damage-classes for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("move-damage-class", id_or_name, **kwargs)


def move_learn_method(id_or_name, **kwargs):
    """Quick move-learn-method lookup.

    See https://pokeapi.co/docsv2/#move-learn-methods for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("move-learn-method", id_or_name, **kwargs)


def move_target(id_or_name, **kwargs):
    """Quick move-target lookup.

    See https://pokeapi.co/docsv2/#move-targets for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("move-target", id_or_name, **kwargs)


def location(id_, **kwargs):
    """Quick location lookup.

    See https://pokeapi.co/docsv2/#locations for attributes and more detailed
    information.

    :param id_: id of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("location", id_, **kwargs)


def location_area(id_, **kwargs):
    """Quick location-area lookup.

    See https://pokeapi.co/docsv2/#location-areas for attributes and more
    detailed information.

    :param id_: id of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("location-area", id_, **kwargs)


def pal_park_area(id_or_name, **kwargs):
    """Quick pal-park-area lookup.

    See https://pokeapi.co/docsv2/#pal-park-areas for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("pal-park-area", id_or_name, **kwargs)


def region(id_or_name, **kwargs):
    """Quick region lookup.

    See https://pokeapi.co/docsv2/#regions for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("region", id_or_name, **kwargs)


def ability(id_or_name, **kwargs):
    """Quick ability lookup.

    See https://pokeapi.co/docsv2/#abilities for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("ability", id_or_name, **kwargs)


def characteristic(id_, **kwargs):
    """Quick characteristic lookup.

    See https://pokeapi.co/docsv2/#characteristics for attributes and more
    detailed information.

    :param id_: id of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("characteristic", id_, **kwargs)


def egg_group(id_or_name, **kwargs):
    """Quick egg-group lookup.

    See https://pokeapi.co/docsv2/#egg-groups for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("egg-group", id_or_name, **kwargs)


def gender(id_or_name, **kwargs):
    """Quick gender lookup.

    See https://pokeapi.co/docsv2/#genders for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("gender", id_or_name, **kwargs)


def growth_rate(id_or_name, **kwargs):
    """Quick growth-rate lookup.

    See https://pokeapi.co/docsv2/#growth-rates for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("growth-rate", id_or_name, **kwargs)


def nature(id_or_name, **kwargs):
    """Quick nature lookup.

    See https://pokeapi.co/docsv2/#natures for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("nature", id_or_name, **kwargs)


def pokeathlon_stat(id_or_name, **kwargs):
    """Quick pokeathlon-stat lookup.

    See https://pokeapi.co/docsv2/#pokeathlon-stats for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("pokeathlon-stat", id_or_name, **kwargs)


def pokemon(id_or_name, **kwargs):
    """Quick pokemon lookup.

    See https://pokeapi.co/docsv2/#pokemon for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """

    def get_location_area_encounters(val):
        params = val.split("/")[-3:]
        params[1] = int(params[1])
        return params

    return APIResource(
        "pokemon",
        id_or_name,
        custom={"location_area_encounters": get_location_area_encounters},
        **kwargs
    )


def pokemon_color(id_or_name, **kwargs):
    """Quick pokemon-color lookup.

    See https://pokeapi.co/docsv2/#pokemon-colors for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("pokemon-color", id_or_name, **kwargs)


def pokemon_form(id_or_name, **kwargs):
    """Quick pokemon-form lookup.

    See https://pokeapi.co/docsv2/#pokemon-forms for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("pokemon-form", id_or_name, **kwargs)


def pokemon_habitat(id_or_name, **kwargs):
    """Quick pokemon-habitat lookup.

    See https://pokeapi.co/docsv2/#pokemon-habitats for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("pokemon-habitat", id_or_name, **kwargs)


def pokemon_shape(id_or_name, **kwargs):
    """Quick pokemon-shape lookup.

    See https://pokeapi.co/docsv2/#pokemon-shapes for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("pokemon-shape", id_or_name, **kwargs)


def pokemon_species(id_or_name, **kwargs):
    """Quick pokemon-species lookup.

    See https://pokeapi.co/docsv2/#pokemon-species for attributes and more
    detailed information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """

    def get_evolution_chain(val):
        params = val["url"].split("/")[-3:-1]
        params[1] = int(params[1])
        return params

    return APIResource(
        "pokemon-species",
        id_or_name,
        custom={"evolution_chain": get_evolution_chain},
        **kwargs
    )


def stat(id_or_name, **kwargs):
    """Quick stat lookup.

    See https://pokeapi.co/docsv2/#stats for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("stat", id_or_name, **kwargs)


def type_(id_or_name, **kwargs):
    """Quick type lookup.

    See https://pokeapi.co/docsv2/#types for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("type", id_or_name, **kwargs)


def language(id_or_name, **kwargs):
    """Quick language lookup.

    See https://pokeapi.co/docsv2/#languages for attributes and more detailed
    information.

    :param id_or_name: id or name of the resource to lookup
    :return: NamedAPIResource with the appropriate data
    """
    return APIResource("language", id_or_name, **kwargs)


def sprite(sprite_type, sprite_id, **kwargs):
    return SpriteResource(sprite_type, sprite_id, **kwargs)



# -*- coding: utf-8 -*-

import shelve
import unittest
from unittest.mock import patch

from hypothesis import assume, given
from hypothesis.strategies import dictionaries, integers, none, sampled_from, text
from requests.exceptions import HTTPError

from pokebase import api, cache
from pokebase.cache import save, set_cache
from pokebase.common import ENDPOINTS, cache_uri_build


class TestFunction__call_api(unittest.TestCase):

    # _call_api(endpoint, resource_id)

    def setUp(self):
        set_cache('testing')

    @given(endpoint=sampled_from(ENDPOINTS),
           resource_id=(integers(min_value=1)))
    @patch('pokebase.api.requests.get')
    def testArgs(self, mock_get, endpoint, resource_id):

        mock_get.return_value.json.return_value = {'id': resource_id}

        self.assertEqual(api._call_api(endpoint, resource_id)['id'],
                         resource_id)

    @given(endpoint=text(),
           resource_id=(integers(min_value=1)))
    def testArg_endpoint_Text(self, endpoint, resource_id):
        with self.assertRaises(ValueError):
            api._call_api(endpoint, resource_id)

    @given(endpoint=sampled_from(ENDPOINTS),
           resource_id=none())
    @patch('pokebase.api.requests.get')
    def testArg_resource_id_None(self, mock_get, endpoint, resource_id):

        mock_get.return_value.json.return_value = {'count': 100, 'results': ['some', 'reults']}

        self.assertIsNotNone(api._call_api(endpoint, resource_id).get('count'))

    @given(endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1),
           subresource=text())
    @patch('pokebase.api.requests.get')
    def testArg_subresource_Text(self, mock_get, endpoint, resource_id, subresource):
        mock_get.return_value.json.return_value = {'version_details': 'foo'}

        self.assertIsNotNone(api._call_api(endpoint, resource_id, subresource).get('version_details'))

    @given(endpoint=sampled_from(ENDPOINTS),
           resource_id=(integers(min_value=1)))
    @patch('pokebase.api.requests.get')
    def testEnv_ErrorResponse(self, mock_get, endpoint, resource_id):
        mock_get.return_value.raise_for_status.side_effect = HTTPError()

        with self.assertRaises(HTTPError):
            api._call_api(endpoint, resource_id)


class TestFunction_get_data(unittest.TestCase):

    # get_data(endpoint, resource_id)

    def setUp(self):
        set_cache('testing')

    @given(data=dictionaries(text(), text()),
           endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1))
    def testArgs_GettingCachedData(self, data, endpoint, resource_id):
        assume(data != dict())
        # save some data to the cache
        save(data, endpoint, resource_id)
        self.assertEqual(data, api.get_data(endpoint, resource_id))

    @given(data=dictionaries(text(), text()),
           endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1))
    @patch('pokebase.api.requests.get')
    def testArgs_GettingNoncachedData(self, mock_get, data, endpoint, resource_id):

        mock_get.return_value.json.return_value = data

        # assert that the data is not in the cache
        with shelve.open(cache.API_CACHE) as cache_file:
            key = cache_uri_build(endpoint, resource_id)
            if key in cache_file.keys():
                del cache_file[key]

        self.assertEqual(data, api.get_data(endpoint, resource_id))

    @given(endpoint=sampled_from(ENDPOINTS),
           resource_id=integers(min_value=1),
           subresource=text())
    @patch('pokebase.api.requests.get')
    def testArg_subresource_Text(self, mock_get, endpoint, resource_id, subresource):
        mock_get.return_value.json.return_value = {'version_details': 'foo'}

        self.assertIsNotNone(api.get_data(endpoint, resource_id, subresource).get('version_details'))

    @given(endpoint=text(),
           resource_id=integers(min_value=1))
    def testArg_endpoint_Text(self, endpoint, resource_id):
        with self.assertRaises(ValueError):
            api.get_data(endpoint, resource_id)

    @given(endpoint=sampled_from(ENDPOINTS),
           resource_id=text())
    def testArg_resource_id_Text(self, endpoint, resource_id):
        with self.assertRaises(ValueError):
            api.get_data(endpoint, resource_id)



import setuptools


def readme_text():
    with open('README.md') as f:
        return f.read()


setuptools.setup(
    name='pokebase',
    packages=['pokebase'],
    version='1.4.0',
    description='A Python wrapper for the friendly PokeAPI database',
    long_description=readme_text(),
    author='Greg Hilmes',
    author_email='99hilmes.g@gmail.com',
    url='https://github.com/PokeAPI/pokebase',
    keywords=['database', 'pokemon', 'wrapper'],
    install_requires=['requests'],
    license='BSD License',
    requires_python = ">=3.8",
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12'
    ]
)



# -*- coding: utf-8 -*-

from .interface import APIMetadata, APIResource, APIResourceList, SpriteResource
from .loaders import *

__all__ = [
    "APIResource",
    "APIMetadata",
    "APIResourceList",
    "SpriteResource",
    "ability",
    "berry",
    "berry_firmness",
    "berry_flavor",
    "characteristic",
    "contest_effect",
    "contest_type",
    "egg_group",
    "encounter_condition",
    "encounter_condition_value",
    "encounter_method",
    "evolution_chain",
    "evolution_trigger",
    "gender",
    "generation",
    "growth_rate",
    "item",
    "item_attribute",
    "item_category",
    "item_fling_effect",
    "item_pocket",
    "language",
    "location",
    "location_area",
    "machine",
    "move",
    "move_ailment",
    "move_battle_style",
    "move_category",
    "move_damage_class",
    "move_learn_method",
    "move_target",
    "nature",
    "pal_park_area",
    "pokeathlon_stat",
    "pokedex",
    "pokemon",
    "pokemon_color",
    "pokemon_form",
    "pokemon_habitat",
    "pokemon_shape",
    "pokemon_species",
    "region",
    "stat",
    "super_contest_effect",
    "type_",
    "version",
    "version_group",
]
